@@ -1,7 +1,7 @@
 #!/bin/env python3
 
 # mcctl: A Minecraft Server Management Utility written in Python
-# Copyright (C) 2020 Matthias Cotting
+# Copyright (C) 2021 Matthias Cotting
 
 # This file is part of mcctl.
 
@@ -19,7 +19,6 @@
 # along with mcctl. If not, see <http://www.gnu.org/licenses/>.
 
 
-__version__ = "0.3.1"
-
 from mcctl.__config__ import CFGVARS  # noqa: F401
-from mcctl import common, config, proc, service, storage, visuals, web  # noqa: F401
+from mcctl import common, config, plugin, proc, service, storage, visuals, web  # noqa: F401
+__version__ = "0.4.0"
