@@ -1,7 +1,7 @@
 #!/bin/env python3
 
 # mcctl: A Minecraft Server Management Utility written in Python
-# Copyright (C) 2020 Matthias Cotting
+# Copyright (C) 2021 Matthias Cotting
 
 # This file is part of mcctl.
 
@@ -24,7 +24,7 @@ import inspect
 import argparse as ap
 from typing import Callable
 from mcctl.__config__ import LOGIN_USER, read_cfg, write_cfg
-from mcctl import proc, storage, service, web, common, CFGVARS, __version__
+from mcctl import proc, storage, plugin, perms, web, common, CFGVARS, __version__
 
 
 def get_permlevel(args: ap.Namespace, elevation: dict) -> dict:
@@ -48,22 +48,22 @@ def get_permlevel(args: ap.Namespace, elevation: dict) -> dict:
         "root": "root"
     }
 
-    perms = {"usr": users.get(elevation.get("default"))}
+    permissions = {"usr": users.get(elevation.get("default"))}
     conditions = elevation.get("on_cond")
     cond_match = not bool(conditions)
     if conditions:
         kwargs = vars(args)
         for key, val in conditions.items():
-            if kwargs[key] == val:
+            if key in kwargs and kwargs[key] == val:
                 cond_match = True
                 break
 
     change_to = elevation.get("change_to")
     if cond_match and change_to:
-        perms["usr"] = users.get(change_to)
+        permissions["usr"] = users.get(change_to)
         if not elevation.get("change_fully", False):
-            perms["eusr"] = users.get(elevation.get("default"))
-    return perms
+            permissions["eusr"] = users.get(elevation.get("default"))
+    return permissions
 
 
 def apply_permlevel(permlevel: dict) -> None:
@@ -74,11 +74,11 @@ def apply_permlevel(permlevel: dict) -> None:
             - "usr": The user to elevate to via sudo.
             - "eusr" (optional): The user of which the EIDs are set.
     """
-    proc.elevate(permlevel.get('usr'))
+    perms.elevate(permlevel.get('usr'))
     demote_user = permlevel.get('eusr')
     if demote_user:
         user_ids = proc.get_ids(demote_user)
-        proc.run_as(*user_ids)
+        perms.set_eids(*user_ids)
 
 
 def filter_args(unfiltered_kwargs: dict, func: Callable) -> dict:
@@ -109,47 +109,63 @@ def get_parser() -> ap.ArgumentParser:
     """Parse Arguments from the Command Line input and returns the converted Values.
 
     Returns:
-        argparse.Namespace: All Arguments, parsed.
+        ArgumentParser: A parser with all arguments.
 
     Raises:
         argparse.ArgumentTypeError: Raised when the parameters given cannot be parsed correctly.
     """
     def check_type_id(value: str) -> str:
-        test_type_id = re.compile(
-            r'(.+:)+.+|https?: \/\/(-\.)?([ ^\s /?\.#-]+\.?)+(/[^\s]*)?$')
-        if not test_type_id.search(value):
+        if not (re.search(r'^[a-z0-9\._-]+(:[a-z0-9\._-]+){1,2}$', value) or web.is_url(value)):
             raise ap.ArgumentTypeError(
                 "must be in the form '<TYPE>:<VERSION>:<BUILD>' or URL.")
         return value
 
     def check_strict_type_id(value: str) -> str:
-        test_type_id = re.compile(r'(.+:)+.+|all')
-        if not test_type_id.search(value):
+        if not re.search(r'^[a-z0-9\._-]+(:[a-z0-9\._-]+){1,2}$|^all$', value):
             raise ap.ArgumentTypeError(
                 "must be in the form '<TYPE>:<VERSION>:<BUILD>' or 'all'.")
         return value
 
     def check_mem(value: str) -> str:
-        test_mem = re.compile(r'^[0-9]+[KMG]$')
-        if not test_mem.search(value):
+        if not re.search(r'^[0-9]*[1-9]+[0-9]*[KMG]$', value):
             raise ap.ArgumentTypeError("Must be in Format <NUMBER>{K,M,G}.")
         return value
 
-    default_err_template = "{args.action} instance '{args.instance}'"
-    default_elev = {"default": "server_user"}
-    default_semi_elev = {"default": "server_user", "change_to": "root"}
+    def autocomplete_instance(value: str) -> str:
+        base = storage.get_instance_path(bare=True)
+        try:
+            matches = [x.name for x in base.iterdir() if x.name.startswith(value)]
+        except OSError:
+            return value
+        if len(matches) > 1:
+            raise ap.ArgumentTypeError(f"Instance Name '{value}' is ambiguous.")
+        elif len(matches) < 1:
+            return value
+        return matches[0]
 
-    parser = ap.ArgumentParser("mcctl", description="Manage, configure, create multiple Minecraft servers in a docker-like fashion.\n"
+    default_err_template = "{args.action} instance '{args.instance}'"
+    no_elev = {"default": "server_user"}
+    semi_elev = {"default": "server_user", "change_to": "root"}
+    root_elev = {"default": "login_user", "change_to": "root", "change_fully": True}
+    re_start_elev = {"default": "server_user",
+                     "change_to": "root",
+                     "on_cond": {'restart': True, 'start': True}}
+
+    parser = ap.ArgumentParser("mcctl", description="Manage, configure, and create multiple Minecraft servers easily with a command-line interface.\n"
                                f"Version: {__version__}",
                                formatter_class=ap.ArgumentDefaultsHelpFormatter)
     parser.add_argument("-v", "--verbose", action='store_true',
                         help="Enable verbose/debugging output.")
     parser.set_defaults(err_template=default_err_template,
-                        elevation=default_elev)
+                        elevation=no_elev)
 
     subparsers = parser.add_subparsers(title="actions", dest="action")
     subparsers.required = True
 
+    force_parser = ap.ArgumentParser(add_help=False)
+    force_parser.add_argument("-f", "--force", action="store_true",
+                              help="Proceed without a prompt.")
+
     type_id_parser = ap.ArgumentParser(
         add_help=False, formatter_class=ap.RawTextHelpFormatter)
     type_id_parser.add_argument(
@@ -162,13 +178,21 @@ def get_parser() -> ap.ArgumentParser:
               "Versions: e.g. '1.15.2', 'latest'\n"
               "Build (only for paper): e.g. '122', 'latest'\n"))
 
-    instance_name_parser = ap.ArgumentParser(add_help=False)
-    instance_name_parser.add_argument(
-        "instance", metavar="INSTANCE_ID", help="Instance Name of the Minecraft Server.")
+    # Auto-Completing Instance Name
+    existing_instance_parser = ap.ArgumentParser(add_help=False)
+    existing_instance_parser.add_argument(
+        "instance", metavar="INSTANCE_ID", type=autocomplete_instance,
+        help="Instance Name of the Minecraft Server.")
+    # New Instance Name
+    instance_parser = ap.ArgumentParser(add_help=False)
+    instance_parser.add_argument(
+        "instance", metavar="INSTANCE_ID",
+        help="Instance Name of the Minecraft Server.")
     # Optional Instance Name
     instance_subfolder_parser = ap.ArgumentParser(add_help=False)
     instance_subfolder_parser.add_argument(
-        "instance_subfolder", metavar="INSTANCE/SUBFOLDER", nargs="?", help="Instance Name or Subpath in Instance Files, e.g. INSTANCE/world.")
+        "instance_subfolder", metavar="INSTANCE/SUBFOLDER", nargs="?", type=autocomplete_instance,
+        help="Instance Name or Subpath in Instance Files, e.g. INSTANCE/world.")
 
     message_parser = ap.ArgumentParser(add_help=False)
     message_parser.add_argument(
@@ -183,26 +207,21 @@ def get_parser() -> ap.ArgumentParser:
         "-m", "--memory", type=check_mem, help="Memory Allocation for the Server in {K,M,G}Bytes, e.g. 2G, 1024M.")
 
     parser_attach = subparsers.add_parser(
-        "attach", parents=[instance_name_parser], help="Attach to the Console of the Minecraft Instance.")
+        "attach", parents=[existing_instance_parser], help="Attach to the Console of the Minecraft Instance.")
     parser_attach.set_defaults(
         func=proc.attach, err_template="attach to '{args.instance}'")
 
     parser_config = subparsers.add_parser(
-        "config", parents=[instance_name_parser, restart_parser, memory_parser], help="Configure/Change Files of a Minecraft Server Instance.")
+        "config", parents=[existing_instance_parser, restart_parser, memory_parser], help="Configure/Change Files of a Minecraft Server Instance.")
     parser_config.add_argument(
         "-e", "--edit", nargs="+", dest="edit_paths", metavar="FILE", help="Edit a File in the Instance Folder interactively.")
     parser_config.add_argument(
         "-p", "--properties", nargs="+", help="Change server.properties options, e.g. server-port=25567 'motd=My new and cool Server'.")
-    parser_config.set_defaults(
-        func=common.configure, err_template="configure '{args.instance}'", editor=CFGVARS.get('user', 'editor'),
-        elevation={
-            "default": "server_user",
-            "change_to": "root",
-            "on_cond": {'restart': True}
-        })
+    parser_config.set_defaults(func=common.configure, err_template="configure '{args.instance}'",
+                               editor=CFGVARS.get('user', 'editor'), elevation=re_start_elev)
 
     parser_create = subparsers.add_parser(
-        "create", parents=[instance_name_parser, type_id_parser, memory_parser], help="Create a new Server Instance.", formatter_class=ap.RawTextHelpFormatter)
+        "create", parents=[instance_parser, type_id_parser, memory_parser], help="Create a new Server Instance.")
     parser_create.add_argument(
         "-s", "--start", action='store_true', help="Start the Server after creation, persistent enabled.")
     parser_create.add_argument(
@@ -216,36 +235,64 @@ def get_parser() -> ap.ArgumentParser:
         })
 
     parser_exec = subparsers.add_parser(
-        "exec", parents=[instance_name_parser], help="Execute a command in the Console of the Instance.")
+        "exec", parents=[existing_instance_parser], help="Execute a command in the Console of the Instance.")
     parser_exec.add_argument("command", nargs="+",
                              help="Command to execute in the Server Console.")
     parser_exec.set_defaults(
         func=proc.mc_exec, err_template="execute command on {args.instance}")
 
     parser_export = subparsers.add_parser(
-        "export", parents=[instance_name_parser], help="Export an Instance to a zip File.")
+        "export", parents=[existing_instance_parser], help="Export an Instance to a zip File.")
     parser_export.add_argument(
         "-c", "--compress", action='store_true', help="Compress the Archive.")
     parser_export.add_argument(
         "-w", "--world-only", action='store_true', help="Only export World Data.")
-    parser_export.set_defaults(
-        func=storage.export, elevation=default_semi_elev)
+    parser_export.set_defaults(func=storage.export, elevation=semi_elev)
+
+    parser_import = subparsers.add_parser(
+        "import", help="Import an Instance from a zip File.")
+    parser_import.add_argument(
+        "zip_path", metavar="ZIPFILE", help="The Path of the archived Server.")
+    parser_import.add_argument(
+        "-i", "--instance", help="Instance Name of the Minecraft Server.")
+    parser_import.add_argument(
+        "-w", "--world-only", action='store_true', help="Only import World Data (the instance must already exist).")
+    parser_import.set_defaults(err_template="{args.action} '{args.zip_path}'",
+                               func=storage.mc_import, elevation=root_elev)
+
+    parser_logs = subparsers.add_parser(
+        "logs", parents=[existing_instance_parser], help="Read the Logs of a Server.")
+    parser_logs.add_argument(
+        "-n", "--lines", dest="limit", type=int, default=0, help="Limit the line output count to n.")
+    parser_logs.set_defaults(
+        func=storage.logs, err_template="read logs of '{args.instance}'")
 
     parser_inspect = subparsers.add_parser(
-        "inspect", parents=[instance_name_parser], help="Inspect the Log of a Server.")
-    parser_inspect.add_argument(
-        "-n", "--lines", dest="limit", type=int, default=0, help="Limit the line output count to n.")
+        "inspect", parents=[existing_instance_parser],
+        help="Get any information you can think of about the server in json format.")
     parser_inspect.set_defaults(
-        func=storage.inspect, err_template="{args.action} logs of '{args.instance}'")
+        func=common.inspect, err_template="{args.action} {args.instance}")
+
+    parser_install = subparsers.add_parser(
+        "install", parents=[existing_instance_parser, restart_parser], formatter_class=ap.RawTextHelpFormatter,
+        help="Install or update a server Plugin from a local Path/URL/Zip File.\n"
+             "Old plugins closely matching the new name can optionally be deleted."
+    )
+    parser_install.add_argument("sources", metavar="LOCAL_PATH_OR_URL", nargs="+",
+                                help="Paths or URLs which point to Plugins or zip Files.")
+    parser_install.add_argument("-a", "--autoupgrade", action='store_true',
+                                help="Upgrade plugin and remove previous versions by name (interactive).")
+    parser_install.set_defaults(func=plugin.install, elevation=root_elev,
+                                err_template="{args.action} plugins on {args.instance}")
 
     parser_list = subparsers.add_parser(
-        "ls", help="List Instances, installed Versions, etc.")
-    parser_list.add_argument("what", metavar="WHAT", nargs="?", choices=[
-        "instances", "jars"], default="instances", help="What Type (instnaces/jars) return.")
+        "ls", help="List Instances, installed Versions, Plugins, etc.")
+    parser_list.add_argument("what", metavar="WHAT", nargs="?", choices=("plugins", "instances", "jars"),
+                             default="instances", help="What Type (instnaces/jars) to return.")
     parser_list.add_argument("-f", "--filter", dest="filter_str",
                              default='', help="Filter by Version or Instance Name, etc.")
     parser_list.set_defaults(
-        func=common.mc_ls, err_template="list {args.what}")
+        func=common.mc_ls, err_template="{args.action} {args.what}")
 
     parser_pull = subparsers.add_parser(
         "pull", parents=[type_id_parser], help="Pull a Server .jar-File from the Internet.")
@@ -253,23 +300,23 @@ def get_parser() -> ap.ArgumentParser:
         func=web.pull, err_template="{args.action} Server Type '{args.source}'")
 
     parser_rename = subparsers.add_parser(
-        "rename", parents=[instance_name_parser], help="Rename a Server Instance.")
+        "rename", parents=[existing_instance_parser], help="Rename a Server Instance.")
     parser_rename.add_argument(
         "new_name", metavar="NEW_NAME", help="The new Name of the Server Instance.")
     parser_rename.set_defaults(func=common.rename)
 
     parser_restart = subparsers.add_parser(
-        "restart", parents=[instance_name_parser, message_parser], help="Restart a Server Instance.")
+        "restart", parents=[existing_instance_parser, message_parser], help="Restart a Server Instance.")
     parser_restart.set_defaults(
-        func=service.notified_set_status, elevation=default_semi_elev)
+        func=common.notified_set_status, elevation=semi_elev)
 
     parser_remove = subparsers.add_parser(
-        "rm", parents=[instance_name_parser], help="Remove a Server Instance.")
+        "rm", parents=[existing_instance_parser, force_parser], help="Remove a Server Instance.")
     parser_remove.set_defaults(
         func=storage.remove, err_template="remove '{args.instance}'")
 
     parser_remove_jar = subparsers.add_parser(
-        "rmj", help="Remove a cached Server Binary.")
+        "rmj", parents=[force_parser], help="Remove a cached Server Binary.")
     parser_remove_jar.add_argument(
         "source", metavar="TYPEID", type=check_strict_type_id,
         help=("Type ID in '<TYPE>:<VERSION>:<BUILD>' format.\n"
@@ -279,28 +326,36 @@ def get_parser() -> ap.ArgumentParser:
         func=storage.remove_jar, err_template="remove .jar File '{args.source}'")
 
     parser_start = subparsers.add_parser(
-        "start", parents=[instance_name_parser], help="Start a Server Instance.")
+        "start", parents=[existing_instance_parser], help="Start a Server Instance.")
     parser_start.add_argument("-p", "--persistent", action='store_true',
                               help="Start even after Reboot.")
     parser_start.set_defaults(
-        func=service.notified_set_status, elevation=default_semi_elev)
+        func=common.notified_set_status, elevation=semi_elev)
+
+    parser_status = subparsers.add_parser(
+        "status", parents=[existing_instance_parser], help="Get extensive Information about the Server Instance.")
+    parser_status.set_defaults(
+        func=common.mc_status, err_template="retrieve {args.action} of '{args.instance}'", elevation=semi_elev)
 
     parser_stop = subparsers.add_parser(
-        "stop", parents=[instance_name_parser, message_parser], help="Stop a Server Instance.")
+        "stop", parents=[existing_instance_parser, message_parser], help="Stop a Server Instance.")
     parser_stop.add_argument("-p", "--persistent", action='store_true',
                              help="Do not start again after Reboot.")
     parser_stop.set_defaults(
-        func=service.notified_set_status, elevation=default_semi_elev)
+        func=common.notified_set_status, elevation=semi_elev)
+
+    parser_uninstall = subparsers.add_parser(
+        "uninstall", parents=[existing_instance_parser, restart_parser, force_parser],
+        help="Uninstall a server Plugin by File Name."
+    )
+    parser_uninstall.add_argument("plugins", metavar="PLUGIN_NAME", nargs="+",
+                                  help="Plugin File names in the plugins folder of the instance.")
+    parser_uninstall.set_defaults(func=plugin.uninstall, elevation=re_start_elev,
+                                  err_template="{args.action} plugins on {args.instance}")
 
     parser_update = subparsers.add_parser(
-        "update", parents=[instance_name_parser, type_id_parser, restart_parser], help="Update a Server Instance.")
-    parser_update.set_defaults(
-        func=common.update,
-        elevation={
-            "default": "server_user",
-            "change_to": "root",
-            "on_cond": {'restart': True}
-        })
+        "update", parents=[existing_instance_parser, type_id_parser, restart_parser], help="Update a Server Instance.")
+    parser_update.set_defaults(func=common.update, elevation=re_start_elev)
 
     parser_shell = subparsers.add_parser(
         "shell", parents=[instance_subfolder_parser], help="Use a Shell to interactively edit a Server Instance.")
@@ -340,7 +395,6 @@ def main() -> None:
             raise
         print(f"Process Elevation failed: {ex}")
         sys.exit(1)
-
     safe_kwargs = filter_args(vars(args), args.func)
     try:
         args.func(**safe_kwargs)
