@@ -1,7 +1,7 @@
 #!/bin/env python3
 
 # mcctl: A Minecraft Server Management Utility written in Python
-# Copyright (C) 2020 Matthias Cotting
+# Copyright (C) 2021 Matthias Cotting
 
 # This file is part of mcctl.
 
@@ -20,15 +20,11 @@
 
 import shlex
 import time
-import os
-import sys
 import subprocess as sproc
-from typing import Callable
-from contextlib import contextmanager
 from pathlib import Path
 from pwd import getpwnam
-from mcctl import CFGVARS, storage, service, common
-from mcctl.visuals import compute
+from mcctl import storage, service, status, perms
+from mcctl.visuals import compute, clear
 
 
 def attach(instance: str) -> None:
@@ -39,11 +35,13 @@ def attach(instance: str) -> None:
     Arguments:
         instance (str): The name of the instance.
     """
-    if not service.is_active(instance):
+    unit = service.get_unit(instance)
+    if not service.is_active(unit):
         raise OSError("The Server is not running.")
     cmd = shlex.split(f"screen -r mc-{instance}")
     proc = sproc.Popen(cmd)
     proc.wait()
+    clear()
 
 
 def shell(instance_subfolder: str, shell_path: Path) -> None:
@@ -77,7 +75,7 @@ def edit(file_path: Path, editor: str) -> None:
         file_path (Path): The file to be edited in the Editor.
     """
     cmd = shlex.split(f"{editor} '{file_path}'")
-    proc = sproc.Popen(cmd, preexec_fn=demote())  # nopep8 pylint: disable=subprocess-popen-preexec-fn
+    proc = sproc.Popen(cmd, preexec_fn=perms.demote())  # nopep8 pylint: disable=subprocess-popen-preexec-fn
     proc.wait()
 
 
@@ -99,9 +97,10 @@ def mc_exec(instance: str, command: list, pollrate: float = 0.2, max_retries: in
         max_retries (int): The amount of retries when no lines have been pushed to console. (default: {25})
         max_flush_retries (int): The amount of retries when some lines have been pushed to console. (default: {10})
     """
-    if not service.is_active(instance):
+    unit = service.get_unit(instance)
+    if not service.is_active(unit):
         raise OSError("The Server is not running.")
-    elif not common.is_ready(instance):
+    elif not status.is_ready(instance):
         raise ConnectionError("The Server is starting up.")
 
     log_path = storage.get_instance_path(instance) / "logs/latest.log"
@@ -113,7 +112,7 @@ def mc_exec(instance: str, command: list, pollrate: float = 0.2, max_retries: in
         # Use ^U^Y to cut and paste Text already in the Session
         cmd = shlex.split(
             f"screen -p 0 -S mc-{instance} -X stuff '^U{jar_cmd}^M^Y'")
-        proc = sproc.Popen(cmd, preexec_fn=demote())  # nopep8 pylint: disable=subprocess-popen-preexec-fn
+        proc = sproc.Popen(cmd, preexec_fn=perms.demote())  # nopep8 pylint: disable=subprocess-popen-preexec-fn
         proc.wait()
 
         i = 0
@@ -141,62 +140,6 @@ def get_ids(user: str) -> tuple:
     return user_data.pw_uid, user_data.pw_gid
 
 
-def run_as(uid: int, gid: int) -> tuple:
-    """Change the user of the currently running python instance.
-
-    Set the EGID and EUID of the running Python script to the permissions of <as_user>.
-
-    Arguments:
-        uid (int): The Effective User ID that is set.
-        gid (int): The Effective Group ID that is set.
-
-    Retruns:
-        old_ids (tuple): A tuple of the UID and GID that were set before the change.
-    """
-    old_ids = (os.geteuid(), os.getegid())
-
-    os.setegid(gid)
-    os.seteuid(uid)
-
-    return old_ids
-
-
-@contextmanager
-def managed_run_as(uid: int, gid: int) -> None:
-    """Manage the User Context and reset it after execution of the "with"-Block.
-
-    Args:
-        uid (int): The Effective User ID that is set during the "with"-Block.
-        gid (int): The Effective Group ID that is set during the "with"-Block.
-    """
-    old = run_as(uid, gid)
-    try:
-        yield
-    finally:
-        run_as(*old)
-
-
-def demote() -> Callable:
-    """Demote a subprocess. for use in preexec_fn.
-
-    Returns:
-        Callable: Returns a function executed by Popen() before running the external command.
-    """
-    user_name = CFGVARS.get('system', 'server_user')
-    user = getpwnam(user_name)
-
-    def set_ids() -> None:
-        if os.getgid() + os.getuid() == 0:
-            # Set EGID and EUID so that GID and UID can be set correctly.
-            os.setegid(0)
-            os.seteuid(0)
-
-            os.setgid(user.pw_gid)
-            os.setuid(user.pw_uid)
-
-    return set_ids
-
-
 def pre_start(jar_path: Path, watch_file: Path = None, kill_sec: int = 80) -> bool:
     """Prepare the server and lets it create configuration files and such.
 
@@ -214,43 +157,22 @@ def pre_start(jar_path: Path, watch_file: Path = None, kill_sec: int = 80) -> bo
         bool: True: The server stopped as expected. False: The server had to be killed.
     """
     cmd = shlex.split(f"/bin/java -jar {jar_path}")
-    proc = sproc.Popen(cmd, cwd=jar_path.parent, stdout=sproc.PIPE,  # nopep8 pylint: disable=subprocess-popen-preexec-fn
-                       stderr=sproc.PIPE, preexec_fn=demote())
+    proc = sproc.Popen(cmd, cwd=jar_path.parent, stdout=sproc.PIPE,  # pylint: disable=subprocess-popen-preexec-fn
+                       stderr=sproc.PIPE, preexec_fn=perms.demote())
 
     fps = 4
     signaled = False
     success = False
     for i in range(kill_sec * fps + 1):
-        print(f"\r{compute(2)} Setting up config files...", end="")
+        print(f"\r{compute(2)} Setting up config files...", end="", flush=True)
         time.sleep(1 / fps)
-        if not signaled and watch_file is not None and watch_file.is_file():
+        if proc.poll() is not None:
+            success = proc.returncode == 0
+            break
+        elif not signaled and watch_file is not None and watch_file.is_file():
             proc.terminate()
             signaled = True
-        elif i == kill_sec * fps:
+        elif i >= kill_sec * fps:
             proc.kill()
-        elif proc.poll() is not None:
-            success = True
-            break
     print()
     return success
-
-
-def elevate(user: str = "root") -> None:
-    """Replace the current Process with a new one as a different User. Requires sudo.
-
-    Args:
-        user (str, optional): The User that will be switched to. Defaults to "root".
-    """
-    desired_uid = getpwnam(user).pw_uid
-    if os.getuid() == desired_uid:
-        return
-
-    package = sys.modules.get('__main__', {}).__package__
-    if package is None:
-        args = sys.argv
-    else:
-        args = [sys.executable, "-m", package] + sys.argv[1:]
-
-    userargs = ["-u", user] if user != 'root' else []
-    sudoargs = ["sudo"] + userargs + args
-    os.execvp(sudoargs[0], sudoargs)
