@@ -1,7 +1,7 @@
 #!/bin/env python3
 
 # mcctl: A Minecraft Server Management Utility written in Python
-# Copyright (C) 2020 Matthias Cotting
+# Copyright (C) 2021 Matthias Cotting
 
 # This file is part of mcctl.
 
@@ -19,6 +19,7 @@
 # along with mcctl. If not, see <http://www.gnu.org/licenses/>.
 
 from pathlib import Path
+from urllib.parse import urlparse
 import time
 import re
 import hashlib
@@ -78,13 +79,20 @@ def get_paper_download_url(version_tag: str, base_url: str) -> tuple:
         minor = version_tag
     else:
         major, minor = version_tag.split(":", 1)
-    test_url = join_url(base_url, major, minor)
+    version_url = join_url(base_url, major, minor)
     try:
-        resolved_data = rest_get(test_url)
-        resolved_tag = ":".join(resolved_data.values())
-    except Exception as ex:
-        raise LookupError("Server Version not found") from ex
-    return join_url(test_url, "download"), resolved_tag
+        resolved_data = rest_get(version_url)
+    except OSError as ex:
+        raise LookupError(f"Server Version not found: {ex}") from ex
+
+    if "error" in resolved_data.keys():
+        error = resolved_data.get("error").capitalize()
+        raise LookupError(f"{error.capitalize()}")
+
+    # Make sure revision is str and not int
+    resolved_tag = ":".join(str(x) for x in resolved_data.values())
+
+    return join_url(version_url, "download"), resolved_tag
 
 
 def get_spigot_download_url(version_tag: str, base_url: str) -> tuple:
@@ -142,13 +150,17 @@ def get_download_url(server_tag: str) -> tuple:
     Returns:
         tuple: A tuple with the download URL and the complete, resolved Tag
     """
-    assert ":" in server_tag, f"Invalid Server Tag '{server_tag}'"
+    if ":" not in server_tag:
+        raise ValueError(f"Invalid Server Tag '{server_tag}'")
     type_tag, version_tag = server_tag.split(":", 1)
     try:
-        func = SOURCES.get(type_tag, {}).get('func')
-        url, resolved_tag = func(version_tag, SOURCES.get(type_tag).get('url'))
-    except AttributeError:
-        raise ValueError("Unsupported server type") from None
+        source_tools = SOURCES.get(type_tag)
+    except AttributeError as ex:
+        raise ValueError("Unsupported server type.") from ex
+
+    func = source_tools.get('func')
+    base_url = source_tools.get('url')
+    url, resolved_tag = func(version_tag, base_url)
 
     return url, resolved_tag
 
@@ -196,9 +208,22 @@ def download(url: str, dest: Path) -> None:
     Arguments:
         url (str): The target to query.
         dest (Path): The path where to save the recieved file to.
+    Returns:
+        Path: The absolute destination Path, filename included.
     """
+    dest = Path(dest)
     response = req.get(url, stream=True)
-    with open(dest, "wb") as dest_hnd:
+    if dest.is_dir():
+        fdisp = response.headers.get('content-disposition')
+        if fdisp is not None:
+            fname = re.findall(r"filename\*?=(?:.+'')?(.+)", fdisp)[-1]
+            file_dest = dest / fname
+        else:
+            file_dest = dest / url.split("/")[-1]
+    else:
+        file_dest = dest
+
+    with open(file_dest, "wb") as dest_hnd:
         total_length = int(response.headers.get('content-length', 0))
 
         if not total_length:
@@ -211,27 +236,25 @@ def download(url: str, dest: Path) -> None:
                 loaded += len(data)
                 dest_hnd.write(data)
                 elapsed = time.time() - inital
-                progress(loaded, elapsed, total_length)
+                visuals.progress(loaded, elapsed, total_length)
         print()
+    return file_dest.absolute()
 
 
-def progress(current: int, elapsed: float, total: int) -> None:
-    """Print Progress.
+def is_url(url: str) -> bool:
+    """Test if the string is a valid URL.
 
-    Output the progress of the download given blockcount, blocksize and total bytes.
+    Args:
+        url (str): A string that is potentially a URL
 
-    Arguments:
-        downloaded (int): The number of bits recieved.
-        elapsed (int): Elapsed time in seconds.
-        total (int): The size of the complete File.
+    Returns:
+        bool: True if the submitted String is a valid URL.
     """
-    spinner = visuals.SPINNERS[1]
-    chars = spinner.get('chars')
-    char_idx = int((elapsed * spinner.get('fps')) % len(chars))
-
-    percent = current * 100 / total
-    out = f"\r{chars[char_idx]} {percent:3.0f}% {current / 1024 :>{len(str(total // 1024))}.0f}kB / {(total/1024):.0f}kB"
-    print(out, end="")
+    try:
+        seg = urlparse(url)
+    except ValueError:
+        return False
+    return all((seg.scheme, seg.netloc, seg.path))
 
 
 def join_url(base: str, *parts: str) -> str:
@@ -266,11 +289,15 @@ def pull(source: str, literal_url: bool = False) -> tuple:
         url = source
         # Generate artificial Version Tag
         url_hash = hashlib.sha1(url.encode()).hexdigest()
-        tag = f"other/{url_hash[:12]}"
+        tag = f"other:{url_hash[:16]}"
     else:
-        url, tag = get_download_url(source)
+        try:
+            url, tag = get_download_url(source)
+        except LookupError:
+            print("Resolving Tag failed, looking in local cache.")
+            tag = source
 
-    print(f"Pulling version '{tag}'")
+    print(f"Pulling version '{tag}'...")
     dest = storage.get_jar_path(tag)
 
     if not dest.is_file():
